@@ -11,7 +11,7 @@ pub struct Zsh;
 
 impl Generator for Zsh {
     fn file_name(&self, name: &str) -> String {
-        format!("_{}", name)
+        format!("_{name}")
     }
 
     fn generate(&self, cmd: &Command, buf: &mut dyn Write) {
@@ -43,7 +43,11 @@ _{name}() {{
 
 {subcommand_details}
 
-_{name} \"$@\"
+if [ \"$funcstack[1]\" = \"_{name}\" ]; then
+    _{name} \"$@\"
+else
+    compdef _{name} {name}
+fi
 ",
                 name = bin_name,
                 initial_args = get_args_of(cmd, None),
@@ -111,8 +115,8 @@ _{bin_name_underscore}_commands() {{
     all_subcommands.sort();
     all_subcommands.dedup();
 
-    for &(_, ref bin_name) in &all_subcommands {
-        debug!("subcommand_details:iter: bin_name={}", bin_name);
+    for (_, ref bin_name) in &all_subcommands {
+        debug!("subcommand_details:iter: bin_name={bin_name}");
 
         ret.push(format!(
             "\
@@ -153,12 +157,10 @@ fn subcommands_of(p: &Command) -> String {
         let text = format!(
             "'{name}:{help}' \\",
             name = name,
-            help = escape_help(subcommand.get_about().unwrap_or(""))
+            help = escape_help(&subcommand.get_about().unwrap_or_default().to_string())
         );
 
-        if !text.is_empty() {
-            ret.push(text);
-        }
+        ret.push(text);
     }
 
     // The subcommands
@@ -225,16 +227,14 @@ fn get_subcommands_of(parent: &Command) -> String {
     let subcommand_names = utils::subcommands(parent);
     let mut all_subcommands = vec![];
 
-    for &(ref name, ref bin_name) in &subcommand_names {
+    for (ref name, ref bin_name) in &subcommand_names {
         debug!(
-            "get_subcommands_of:iter: parent={}, name={}, bin_name={}",
+            "get_subcommands_of:iter: parent={}, name={name}, bin_name={bin_name}",
             parent.get_name(),
-            name,
-            bin_name,
         );
-        let mut segments = vec![format!("({})", name)];
+        let mut segments = vec![format!("({name})")];
         let subcommand_args = get_args_of(
-            parser_of(parent, &*bin_name).expect(INTERNAL_ERROR_MSG),
+            parser_of(parent, bin_name).expect(INTERNAL_ERROR_MSG),
             Some(parent),
         );
 
@@ -243,7 +243,7 @@ fn get_subcommands_of(parent: &Command) -> String {
         }
 
         // Get the help text of all child subcommands.
-        let children = get_subcommands_of(parser_of(parent, &*bin_name).expect(INTERNAL_ERROR_MSG));
+        let children = get_subcommands_of(parser_of(parent, bin_name).expect(INTERNAL_ERROR_MSG));
 
         if !children.is_empty() {
             segments.push(children);
@@ -280,13 +280,10 @@ esac",
 //
 // Given the bin_name "a b c" and the Command for "a" this returns the "c" Command.
 // Given the bin_name "a b c" and the Command for "b" this returns the "c" Command.
-fn parser_of<'help, 'cmd>(
-    parent: &'cmd Command<'help>,
-    bin_name: &str,
-) -> Option<&'cmd Command<'help>> {
+fn parser_of<'cmd>(parent: &'cmd Command, bin_name: &str) -> Option<&'cmd Command> {
     debug!("parser_of: p={}, bin_name={}", parent.get_name(), bin_name);
 
-    if bin_name == parent.get_bin_name().unwrap_or(&String::new()) {
+    if bin_name == parent.get_bin_name().unwrap_or_default() {
         return Some(parent);
     }
 
@@ -359,7 +356,7 @@ fn get_args_of(parent: &Command, p_global: Option<&Command>) -> String {
 
 // Uses either `possible_vals` or `value_hint` to give hints about possible argument values
 fn value_completion(arg: &Arg) -> Option<String> {
-    if let Some(values) = &arg.get_possible_values() {
+    if let Some(values) = crate::generator::utils::possible_values(arg) {
         if values
             .iter()
             .any(|value| !value.is_hide_set() && value.get_help().is_some())
@@ -375,7 +372,8 @@ fn value_completion(arg: &Arg) -> Option<String> {
                             Some(format!(
                                 r#"{name}\:"{tooltip}""#,
                                 name = escape_value(value.get_name()),
-                                tooltip = value.get_help().map(escape_help).unwrap_or_default()
+                                tooltip =
+                                    escape_help(&value.get_help().unwrap_or_default().to_string()),
                             ))
                         }
                     })
@@ -388,7 +386,7 @@ fn value_completion(arg: &Arg) -> Option<String> {
                 values
                     .iter()
                     .filter(|pv| !pv.is_hide_set())
-                    .map(PossibleValue::get_name)
+                    .map(|n| n.get_name())
                     .collect::<Vec<_>>()
                     .join(" ")
             ))
@@ -428,6 +426,9 @@ fn escape_help(string: &str) -> String {
         .replace('\'', "'\\''")
         .replace('[', "\\[")
         .replace(']', "\\]")
+        .replace(':', "\\:")
+        .replace('$', "\\$")
+        .replace('`', "\\`")
 }
 
 /// Escape value string inside single quotes and parentheses
@@ -435,6 +436,11 @@ fn escape_value(string: &str) -> String {
     string
         .replace('\\', "\\\\")
         .replace('\'', "'\\''")
+        .replace('[', "\\[")
+        .replace(']', "\\]")
+        .replace(':', "\\:")
+        .replace('$', "\\$")
+        .replace('`', "\\`")
         .replace('(', "\\(")
         .replace(')', "\\)")
         .replace(' ', "\\ ")
@@ -448,10 +454,10 @@ fn write_opts_of(p: &Command, p_global: Option<&Command>) -> String {
     for o in p.get_opts() {
         debug!("write_opts_of:iter: o={}", o.get_id());
 
-        let help = o.get_help().map_or(String::new(), escape_help);
+        let help = escape_help(&o.get_help().unwrap_or_default().to_string());
         let conflicts = arg_conflicts(p, o, p_global);
 
-        let multiple = if o.is_multiple_occurrences_set() {
+        let multiple = if let ArgAction::Count | ArgAction::Append = o.get_action() {
             "*"
         } else {
             ""
@@ -462,24 +468,14 @@ fn write_opts_of(p: &Command, p_global: Option<&Command>) -> String {
             Some(val) => val[0].to_string(),
         };
         let vc = match value_completion(o) {
-            Some(val) => format!(":{}:{}", vn, val),
-            None => format!(":{}: ", vn),
-        };
-        let vc = match o.get_num_vals() {
-            Some(num_vals) => vc.repeat(num_vals),
-            None => vc,
+            Some(val) => format!(":{vn}:{val}"),
+            None => format!(":{vn}: "),
         };
+        let vc = vc.repeat(o.get_num_args().expect("built").min_values());
 
         if let Some(shorts) = o.get_short_and_visible_aliases() {
             for short in shorts {
-                let s = format!(
-                    "'{conflicts}{multiple}-{arg}+[{help}]{value_completion}' \\",
-                    conflicts = conflicts,
-                    multiple = multiple,
-                    arg = short,
-                    value_completion = vc,
-                    help = help
-                );
+                let s = format!("'{conflicts}{multiple}-{short}+[{help}]{vc}' \\");
 
                 debug!("write_opts_of:iter: Wrote...{}", &*s);
                 ret.push(s);
@@ -487,14 +483,7 @@ fn write_opts_of(p: &Command, p_global: Option<&Command>) -> String {
         }
         if let Some(longs) = o.get_long_and_visible_aliases() {
             for long in longs {
-                let l = format!(
-                    "'{conflicts}{multiple}--{arg}=[{help}]{value_completion}' \\",
-                    conflicts = conflicts,
-                    multiple = multiple,
-                    arg = long,
-                    value_completion = vc,
-                    help = help
-                );
+                let l = format!("'{conflicts}{multiple}--{long}=[{help}]{vc}' \\");
 
                 debug!("write_opts_of:iter: Wrote...{}", &*l);
                 ret.push(l);
@@ -509,11 +498,11 @@ fn arg_conflicts(cmd: &Command, arg: &Arg, app_global: Option<&Command>) -> Stri
     fn push_conflicts(conflicts: &[&Arg], res: &mut Vec<String>) {
         for conflict in conflicts {
             if let Some(s) = conflict.get_short() {
-                res.push(format!("-{}", s));
+                res.push(format!("-{s}"));
             }
 
             if let Some(l) = conflict.get_long() {
-                res.push(format!("--{}", l));
+                res.push(format!("--{l}"));
             }
         }
     }
@@ -551,23 +540,17 @@ fn write_flags_of(p: &Command, p_global: Option<&Command>) -> String {
     for f in utils::flags(p) {
         debug!("write_flags_of:iter: f={}", f.get_id());
 
-        let help = f.get_help().map_or(String::new(), escape_help);
+        let help = escape_help(&f.get_help().unwrap_or_default().to_string());
         let conflicts = arg_conflicts(p, &f, p_global);
 
-        let multiple = if f.is_multiple_occurrences_set() {
+        let multiple = if let ArgAction::Count | ArgAction::Append = f.get_action() {
             "*"
         } else {
             ""
         };
 
         if let Some(short) = f.get_short() {
-            let s = format!(
-                "'{conflicts}{multiple}-{arg}[{help}]' \\",
-                multiple = multiple,
-                conflicts = conflicts,
-                arg = short,
-                help = help
-            );
+            let s = format!("'{conflicts}{multiple}-{short}[{help}]' \\");
 
             debug!("write_flags_of:iter: Wrote...{}", &*s);
 
@@ -575,13 +558,7 @@ fn write_flags_of(p: &Command, p_global: Option<&Command>) -> String {
 
             if let Some(short_aliases) = f.get_visible_short_aliases() {
                 for alias in short_aliases {
-                    let s = format!(
-                        "'{conflicts}{multiple}-{arg}[{help}]' \\",
-                        multiple = multiple,
-                        conflicts = conflicts,
-                        arg = alias,
-                        help = help
-                    );
+                    let s = format!("'{conflicts}{multiple}-{alias}[{help}]' \\",);
 
                     debug!("write_flags_of:iter: Wrote...{}", &*s);
 
@@ -591,13 +568,7 @@ fn write_flags_of(p: &Command, p_global: Option<&Command>) -> String {
         }
 
         if let Some(long) = f.get_long() {
-            let l = format!(
-                "'{conflicts}{multiple}--{arg}[{help}]' \\",
-                conflicts = conflicts,
-                multiple = multiple,
-                arg = long,
-                help = help
-            );
+            let l = format!("'{conflicts}{multiple}--{long}[{help}]' \\");
 
             debug!("write_flags_of:iter: Wrote...{}", &*l);
 
@@ -605,13 +576,7 @@ fn write_flags_of(p: &Command, p_global: Option<&Command>) -> String {
 
             if let Some(aliases) = f.get_visible_aliases() {
                 for alias in aliases {
-                    let l = format!(
-                        "'{conflicts}{multiple}--{arg}[{help}]' \\",
-                        conflicts = conflicts,
-                        multiple = multiple,
-                        arg = alias,
-                        help = help
-                    );
+                    let l = format!("'{conflicts}{multiple}--{alias}[{help}]' \\");
 
                     debug!("write_flags_of:iter: Wrote...{}", &*l);
 
@@ -629,11 +594,49 @@ fn write_positionals_of(p: &Command) -> String {
 
     let mut ret = vec![];
 
+    // Completions for commands that end with two Vec arguments require special care.
+    // - You can have two Vec args separated with a custom value terminator.
+    // - You can have two Vec args with the second one set to last (raw sets last)
+    //   which will require a '--' separator to be used before the second argument
+    //   on the command-line.
+    //
+    // We use the '-S' _arguments option to disable completion after '--'. Thus, the
+    // completion for the second argument in scenario (B) does not need to be emitted
+    // because it is implicitly handled by the '-S' option.
+    // We only need to emit the first catch-all.
+    //
+    // Have we already emitted a catch-all multi-valued positional argument
+    // without a custom value terminator?
+    let mut catch_all_emitted = false;
+
     for arg in p.get_positionals() {
         debug!("write_positionals_of:iter: arg={}", arg.get_id());
 
-        let cardinality = if arg.is_multiple_values_set() || arg.is_multiple_occurrences_set() {
-            "*:"
+        let num_args = arg.get_num_args().expect("built");
+        let is_multi_valued = num_args.max_values() > 1;
+
+        if catch_all_emitted && (arg.is_last_set() || is_multi_valued) {
+            // This is the final argument and it also takes multiple arguments.
+            // We've already emitted a catch-all positional argument so we don't need
+            // to emit anything for this argument because it is implicitly handled by
+            // the use of the '-S' _arguments option.
+            continue;
+        }
+
+        let cardinality_value;
+        // If we have any subcommands, we'll emit a catch-all argument, so we shouldn't
+        // emit one here.
+        let cardinality = if is_multi_valued && !p.has_subcommands() {
+            match arg.get_value_terminator() {
+                Some(terminator) => {
+                    cardinality_value = format!("*{}:", escape_value(terminator));
+                    cardinality_value.as_str()
+                }
+                None => {
+                    catch_all_emitted = true;
+                    "*:"
+                }
+            }
         } else if !arg.is_required_set() {
             ":"
         } else {
@@ -646,18 +649,43 @@ fn write_positionals_of(p: &Command) -> String {
             name = arg.get_id(),
             help = arg
                 .get_help()
-                .map_or("".to_owned(), |v| " -- ".to_owned() + v)
+                .map(|s| s.to_string())
+                .map(|v| " -- ".to_owned() + &v)
+                .unwrap_or_else(|| "".to_owned())
                 .replace('[', "\\[")
                 .replace(']', "\\]")
                 .replace('\'', "'\\''")
                 .replace(':', "\\:"),
-            value_completion = value_completion(arg).unwrap_or_else(|| "".to_string())
+            value_completion = value_completion(arg).unwrap_or_default()
         );
 
-        debug!("write_positionals_of:iter: Wrote...{}", a);
+        debug!("write_positionals_of:iter: Wrote...{a}");
 
         ret.push(a);
     }
 
     ret.join("\n")
 }
+
+#[cfg(test)]
+mod tests {
+    use crate::shells::zsh::{escape_help, escape_value};
+
+    #[test]
+    fn test_escape_value() {
+        let raw_string = "\\ [foo]() `bar https://$PATH";
+        assert_eq!(
+            escape_value(raw_string),
+            "\\\\\\ \\[foo\\]\\(\\)\\ \\`bar\\ https\\://\\$PATH"
+        )
+    }
+
+    #[test]
+    fn test_escape_help() {
+        let raw_string = "\\ [foo]() `bar https://$PATH";
+        assert_eq!(
+            escape_help(raw_string),
+            "\\\\ \\[foo\\]() \\`bar https\\://\\$PATH"
+        )
+    }
+}
